@@ -1,6 +1,6 @@
 /*
  * sar: report system activity
- * (C) 1999-2008 by Sebastien GODARD (sysstat <at> orange.fr)
+ * (C) 1999-2009 by Sebastien GODARD (sysstat <at> orange.fr)
  *
  ***************************************************************************
  * This program is free software; you can redistribute it and/or modify it *
@@ -52,7 +52,7 @@ int dis = TRUE;
 unsigned int flags = 0;
 char timestamp[2][TIMESTAMP_LEN];
 
-struct stats_sum asum;
+unsigned long avg_count = 0;
 
 /* File header */
 struct file_header file_hdr;
@@ -102,10 +102,10 @@ void usage(char *progname)
 
 	print_usage_title(progname);
 	fprintf(stderr, _("Options are:\n"
-			  "[ -A ] [ -b ] [ -B ] [ -C ] [ -d ] [ -h ] [ -p ] [ -q ] [ -r ] [ -R ]\n"
-			  "[ -S ] [ -t ] [ -u [ ALL ] ] [ -v ] [ -V ] [ -w ] [ -W ] [ -y ]\n"
-			  "[ -I { <int> | SUM | ALL | XALL } ] [ -P { <cpu> | ALL } ]\n"
-			  "[ -n { DEV | EDEV | NFS | NFSD | SOCK | IP | EIP | ICMP | EICMP | TCP | ETCP | UDP | ALL } ]\n"
+			  "[ -A ] [ -b ] [ -B ] [ -C ] [ -d ] [ -h ] [ -m ] [ -p ] [ -q ] [ -r ]\n"
+			  "[ -R ] [ -S ] [ -t ] [ -u [ ALL ] ] [ -v ] [ -V ] [ -w ] [ -W ] [ -y ]\n"
+			  "[ -I { <int> [,...] | SUM | ALL | XALL } ] [ -P { <cpu> [,...] | ALL } ]\n"
+			  "[ -n { <keyword> [,...] | ALL } ]\n"
 			  "[ -o [ <filename> ] | -f [ <filename> ] ]\n"
 			  "[ -i <interval> ] [ -s [ <hh:mm:ss> ] ] [ -e [ <hh:mm:ss> ] ]\n"));
 	exit(1);
@@ -123,15 +123,34 @@ void display_help(char *progname)
 {
 
 	print_usage_title(progname);
-	fprintf(stderr, _("\nMain options and reports:\n"));
+	fprintf(stderr, _("Main options and reports:\n"));
 	fprintf(stderr, _("\t-b\tI/O and transfer rate statistics\n"));
 	fprintf(stderr, _("\t-B\tPaging statistics\n"));
 	fprintf(stderr, _("\t-d\tBlock device statistics\n"));
 	fprintf(stderr, _("\t-I { <int> | SUM | ALL | XALL }\n"
 			  "\t\tInterrupts statistics\n"));
-	fprintf(stderr, _("\t-n { DEV | EDEV | NFS | NFSD | SOCK |\n"
-			  "\t     IP | EIP | ICMP | EICMP | TCP | ETCP | UDP | ALL }\n"
-			  "\t\tNetwork statistics\n"));
+	fprintf(stderr, _("\t-m\tPower management statistics\n"));
+	fprintf(stderr, _("\t-n { <keyword> [,...] | ALL }\n"
+			  "\t\tNetwork statistics\n"
+			  "\t\tKeywords are:\n"
+			  "\t\tDEV\tNetwork interfaces\n"
+			  "\t\tEDEV\tNetwork interfaces (errors)\n"
+			  "\t\tNFS\tNFS client\n"
+			  "\t\tNFSD\tNFS server\n"
+			  "\t\tSOCK\tSockets\t(v4)\n"
+			  "\t\tIP\tIP traffic\t(v4)\n"
+			  "\t\tEIP\tIP traffic\t(v4) (errors)\n"
+			  "\t\tICMP\tICMP traffic\t(v4)\n"
+			  "\t\tEICMP\tICMP traffic\t(v4) (errors)\n"
+			  "\t\tTCP\tTCP traffic\t(v4)\n"
+			  "\t\tETCP\tTCP traffic\t(v4) (errors)\n"
+			  "\t\tUDP\tUDP traffic\t(v4)\n"
+			  "\t\tSOCK6\tSockets\t(v6)\n"
+			  "\t\tIP6\tIP traffic\t(v6)\n"
+			  "\t\tEIP6\tIP traffic\t(v6) (errors)\n"
+			  "\t\tICMP6\tICMP traffic\t(v6)\n"
+			  "\t\tEICMP6\tICMP traffic\t(v6) (errors)\n"
+			  "\t\tUDP6\tUDP traffic\t(v6)\n"));
 	fprintf(stderr, _("\t-q\tQueue length and load average statistics\n"));
 	fprintf(stderr, _("\t-r\tMemory utilization statistics\n"));
 	fprintf(stderr, _("\t-R\tMemory statistics\n"));
@@ -156,8 +175,6 @@ void init_structures(void)
 	
 	for (i = 0; i < 3; i++)
 		memset(&record_hdr[i], 0, RECORD_HEADER_SIZE);
-
-	memset(&asum, 0, STATS_SUM_SIZE);
 }
 
 /*
@@ -220,14 +237,17 @@ void reverse_check_act(unsigned int act_nr)
 
 /*
  ***************************************************************************
- * Fill rectime structure according to time data saved in current
- * structure.
+ * Fill the rectime structure with current record's time, based on current
+ * record's time data saved in file.
+ * The resulting timestamp is expressed in the locale of the file creator
+ * or in the user's own locale depending on whether option -t has been used
+ * or not.
  *
  * IN:
  * @curr	Index in array for current sample statistics.
  ***************************************************************************
 */
-void sar_set_rectime(int curr)
+void sar_get_record_timestamp_struct(int curr)
 {
 	struct tm *ltm;
 
@@ -263,13 +283,13 @@ int check_line_hdr(void)
 	for (i = 0; i < NR_ACT; i++) {
 		if (IS_SELECTED(act[i]->options)) {
 			/* Special processing for activities using a bitmap */
-			if (act[i]->bitmap_size) {
-				if (count_bits(act[i]->bitmap,
-					       BITMAP_SIZE(act[i]->bitmap_size)) > 1) {
+			if (act[i]->bitmap) {
+				if (count_bits(act[i]->bitmap->b_array,
+					       BITMAP_SIZE(act[i]->bitmap->b_size)) > 1) {
 					rc = TRUE;
 				}
 			}
-			else if (*act[i]->nr > 1) {
+			else if (act[i]->nr > 1) {
 				rc = TRUE;
 			}
 			/* Stop now since we have only one selected activity */
@@ -282,7 +302,7 @@ int check_line_hdr(void)
 
 /*
  ***************************************************************************
- * Set timestamp string.
+ * Set current record's timestamp string.
  *
  * IN:
  * @curr	Index in array for current sample statistics.
@@ -292,9 +312,10 @@ int check_line_hdr(void)
  * @cur_time	Timestamp string.
  ***************************************************************************
 */
-void set_timestamp(int curr, char *cur_time, int len)
+void set_record_timestamp_string(int curr, char *cur_time, int len)
 {
-	sar_set_rectime(curr);
+	/* Fill timestamp structure */
+	sar_get_record_timestamp_struct(curr);
 
 	/* Set cur_time date value */
 	strftime(cur_time, len, "%X", &rectime);
@@ -320,7 +341,7 @@ void write_stats_avg(int curr, int read_from_file, unsigned int act_id)
 	static __nr_t cpu_nr = -1;
 	
 	if (cpu_nr < 0)
-		cpu_nr = *act[get_activity_position(act, A_CPU)]->nr;
+		cpu_nr = act[get_activity_position(act, A_CPU)]->nr;
 
 	/* Interval value in jiffies */
 	g_itv = get_interval(record_hdr[2].uptime, record_hdr[curr].uptime);
@@ -341,7 +362,7 @@ void write_stats_avg(int curr, int read_from_file, unsigned int act_id)
 		if ((act_id != ALL_ACTIVITIES) && (act[i]->id != act_id))
 			continue;
 		
-		if (IS_SELECTED(act[i]->options) && (*act[i]->nr > 0)) {
+		if (IS_SELECTED(act[i]->options) && (act[i]->nr > 0)) {
 			/* Display current average activity statistics */
 			if (NEEDS_GLOBAL_ITV(act[i]->options))
 				(*act[i]->f_print_avg)(act[i], 2, curr, g_itv);
@@ -351,8 +372,11 @@ void write_stats_avg(int curr, int read_from_file, unsigned int act_id)
 	}
 
 	if (read_from_file) {
-		/* Reset counters only if we read stats from a system activity file */
-		memset(&asum, 0, STATS_SUM_SIZE);
+		/*
+		 * Reset number of lines printed only if we read stats
+		 * from a system activity file.
+		 */
+		avg_count = 0;
 	}
 }
 
@@ -388,7 +412,7 @@ int write_stats(int curr, int read_from_file, long *cnt, int use_tm_start,
 	static __nr_t cpu_nr = -1;
 
 	if (cpu_nr < 0)
-		cpu_nr = *act[get_activity_position(act, A_CPU)]->nr;
+		cpu_nr = act[get_activity_position(act, A_CPU)]->nr;
 
 	/* Check time (1) */
 	if (read_from_file) {
@@ -399,9 +423,9 @@ int write_stats(int curr, int read_from_file, long *cnt, int use_tm_start,
 	}
 
 	/* Set previous timestamp */
-	set_timestamp(!curr, timestamp[!curr], 16);
+	set_record_timestamp_string(!curr, timestamp[!curr], 16);
 	/* Set current timestamp */
-	set_timestamp(curr,  timestamp[curr],  16);
+	set_record_timestamp_string(curr,  timestamp[curr],  16);
 
 	/* Check if we are beginning a new day */
 	if (use_tm_start && record_hdr[!curr].ust_time &&
@@ -435,7 +459,7 @@ int write_stats(int curr, int read_from_file, long *cnt, int use_tm_start,
 		return 0;
 	}
 
-	(asum.count)++;	/* Nb of lines printed */
+	avg_count++;
 	
 	/* Test stdout */
 	TEST_STDOUT(STDOUT_FILENO);
@@ -445,7 +469,7 @@ int write_stats(int curr, int read_from_file, long *cnt, int use_tm_start,
 		if ((act_id != ALL_ACTIVITIES) && (act[i]->id != act_id))
 			continue;
 
-		if (IS_SELECTED(act[i]->options) && (*act[i]->nr > 0)) {
+		if (IS_SELECTED(act[i]->options) && (act[i]->nr > 0)) {
 			/* Display current activity statistics */
 			if (NEEDS_GLOBAL_ITV(act[i]->options))
 				(*act[i]->f_print)(act[i], !curr, curr, g_itv);
@@ -478,8 +502,8 @@ void write_stats_startup(int curr)
 	record_hdr[!curr].ust_time    = record_hdr[curr].ust_time;
 
 	for (i = 0; i < NR_ACT; i++) {
-		if (IS_SELECTED(act[i]->options) && (*act[i]->nr > 0))
-			memset(act[i]->buf[!curr], 0, act[i]->msize * *act[i]->nr);
+		if (IS_SELECTED(act[i]->options) && (act[i]->nr > 0))
+			memset(act[i]->buf[!curr], 0, act[i]->msize * act[i]->nr);
 	}
 	
 	flags |= S_F_SINCE_BOOT;
@@ -546,7 +570,7 @@ int sar_print_special(int curr, int use_tm_start, int use_tm_end, int rtype, int
 	char cur_time[26];
 	int dp = 1;
 
-	set_timestamp(curr, cur_time, 26);
+	set_record_timestamp_string(curr, cur_time, 26);
 
 	/* The record must be in the interval specified by -s/-e options */
 	if ((use_tm_start && (datecmp(&rectime, &tm_start) < 0)) ||
@@ -601,7 +625,7 @@ void read_sadc_stat_bunch(int curr)
 			PANIC(1);
 		}
 		
-		if (sa_read(act[p]->buf[curr], act[p]->fsize * *act[p]->nr)) {
+		if (sa_read(act[p]->buf[curr], act[p]->fsize * act[p]->nr)) {
 			print_read_error();
 		}
 	}
@@ -651,11 +675,12 @@ void handle_curr_act_stats(int ifd, off_t fpos, int *curr, long *cnt, int *eosaf
 
 	/* Assess number of lines printed */
 	if ((p = get_activity_position(act, act_id)) >= 0) {
-		if (act[p]->bitmap_size) {
-			inc = count_bits(act[p]->bitmap, BITMAP_SIZE(act[p]->bitmap_size));
+		if (act[p]->bitmap) {
+			inc = count_bits(act[p]->bitmap->b_array,
+					 BITMAP_SIZE(act[p]->bitmap->b_size));
 		}
 		else {
-			inc = *act[p]->nr;
+			inc = act[p]->nr;
 		}
 	}
 	if (inc < 0) {
@@ -775,7 +800,7 @@ void read_header_data(void)
 		}
 
 		id_seq[i]   = file_act.id;	/* We necessarily have "i < NR_ACT" */
-		*act[p]->nr = file_act.nr;
+		act[p]->nr  = file_act.nr;
 	}
 
 	while (i < NR_ACT) {
@@ -816,7 +841,7 @@ void read_stats_from_file(char from_file[])
 
 	/* Print report header */
 	print_report_hdr(flags, &rectime, &file_hdr,
-			 *act[get_activity_position(act, A_CPU)]->nr);
+			 act[get_activity_position(act, A_CPU)]->nr);
 
 	/* Read system statistics from file */
 	do {
@@ -840,7 +865,7 @@ void read_stats_from_file(char from_file[])
 				 */
 				read_file_stat_bunch(act, 0, ifd, file_hdr.sa_nr_act,
 						     file_actlst);
-				sar_set_rectime(0);
+				sar_get_record_timestamp_struct(0);
 			}
 		}
 		while ((rtype == R_RESTART) || (rtype == R_COMMENT) ||
@@ -860,7 +885,7 @@ void read_stats_from_file(char from_file[])
 
 		/* Read and write stats located between two possible Linux restarts */
 		for (i = 0; i < NR_ACT; i++) {
-			
+
 			if (!id_seq[i])
 				continue;
 			
@@ -903,6 +928,10 @@ void read_stats_from_file(char from_file[])
 					read_file_stat_bunch(act, curr, ifd, file_hdr.sa_nr_act,
 							     file_actlst);
 				}
+				else if (!eosaf && (rtype == R_COMMENT)) {
+					/* This was a COMMENT record: print it */
+					sar_print_special(curr, tm_start.use, tm_end.use, R_COMMENT, ifd);
+				}
 			}
 			while (!eosaf && (rtype != R_RESTART));
 		}
@@ -954,7 +983,7 @@ void read_stats(void)
 
 	/* Print report header */
 	print_report_hdr(flags, &rectime, &file_hdr,
-			 *act[get_activity_position(act, A_CPU)]->nr);
+			 act[get_activity_position(act, A_CPU)]->nr);
 
 	/* Read system statistics sent by the data collector */
 	read_sadc_stat_bunch(0);
